#include <iostream>
#include <stdarg.h>
#include <vector>
#include <queue>
#include <deque>
#include <limits.h>

using namespace std;

class T{

public:
    int data1;
    int *data2;
// Kopirovaci konstruktor je vytvoren automaticky prekladacem.
// Bohuzel kopiruje automaticky vsechny datove slozky, takze
// v pripade ukazatele zkopiruje jen adresu kterou ukazuje a
// cilova pamet zkopirovana neni.
};

class T2{
public:
    int *data2;
// Vlastni kopirovaci konstruktor
    T2(T2 &t2){
    	data2 = new int(*(t2.data2));
    }
// Pokud predefinuji standardni kopirovaci konstruktor svym,
// prekladac uz sam nevytvori ani prazdny konstruktor.
    T2(){
    	data2 = NULL;
    }

};

// Nasleduje funkce s promennym poctem parametru, neboli s vypustkou
// (ellipse). Prvnim parametrem je pocet argumentu, ktere budou
// zpracovany. Pro zpracovani vypustky je treba vlozit hlavickovy
// soubor <stdarg.h>.
int minimum(int count, ...){
    int tmp;
    // Makro ktere reprezentuje seznam argumentu
    va_list a;
    // Inicializace promenne typu va_list
    va_start(a, count);

    // Pro hledani minima se hodi znat maximalni hodnotu typu integer.
    // Makro INT_MAX je definovano v hlavickovem souboru <limits.h>.
    int ret = INT_MAX;

    for(int i = 0; i < count; ++i){
    	// Makro va_arg ziska z promenne typu va_list dalsi hodnotu
    	// podle typu zadaneho v druhem parametru.
		tmp = va_arg(a, int);
		if(ret > tmp){
			ret = tmp;
		}
    }
    // Je treba jeste list s parametry dealokovat
    va_end(a);
    return ret;
}

int main(){
    T t;

    t.data1 = 1;
    t.data2 = new int(2);

    // Kopirovaci konstruktor - hodnoty instance objektu jsou
    // inicializovany hodnotami instance jineho objektu stejneho typu.
    T t2(t); // ulplne stejne jako: T t2 = t;
    T t3 = t2;
    // Oba predchozi radky jsou naprosto ekvivalentni, pouzije standardni
    // kopirovaci konstruktor vytvoreny prekladacem. Pokud mame datove
    // slozky v objektu jako ukazatele, dojde k jejich zkopirovani.
    // Pokud ukazatel na neco ukazuje, vytvori se melka kopie!

    // Nasleduje vypis, ktery ukazuje problem melke kopie:
    cout << "--- Ukazka melke kopie ---" << endl;
    cout << *(t.data2) << endl;
    *(t2.data2) = 100;
    cout << *(t.data2) << endl;

    // U datoveho typu T2 predchozi problem nenastane, protoze jsme mu
    // vytvorili vlastni kopirovaci konstruktor, ktery vytvari hlubokou\
    // kopii.
    T2 t21;
    t21.data2 = new int(20);
    T2 t22(t21);

    cout << endl << "--- Ukazka hluboke kopie ---" << endl;
    cout << *(t21.data2) << endl;
    *(t22.data2) = 100;
    cout << *(t21.data2) << endl;

    // Ukazka prace s funkci s vypustkou. Nezapomente, ze parametru se
    // zpracuje tolik, kolik je uvedeno v prvnim parametru.
    cout << endl << "--- Ukazka funkci s vypustkou ---" << endl;
    cout << minimum(6, 345, 3, 7, 8, 2, 98) << endl;
    cout << minimum(2, 1, 2) << endl;

    // Ukazka inicializace pomoci konstruktoru
    // Prazdny vektor
    vector<int> v1;
    // Vektor s deseti prvky inicializovanymi na 0
    vector<int> v2(10, 0);
    // Vektor s 20 prvky bez inicializace
    vector<int> v3(20);

    // pridani hodnoty na konec vektoru
    v1.push_back(100);
    cout << endl << "--- Ukazka zakladnich vlastnosti vektoru ---" << endl;
    cout << "v1.size() == " << v1.size() << endl;
    cout << "v1[0] == " << v1[0] << endl << endl;

    v2[3] = 200;
    v2[6] = 300;
    cout << "v2:" << endl;
    for(int i = 0; i < (int) v2.size(); ++i)
    	cout << v2[i] << " ";
    cout << endl << endl;

    // Vypsani vsech prvku vektoru pomoci iteratoru
    cout << "--- Vypsani vektoru v2 pomoci iteratoru ---" << endl;
    for(vector<int>::iterator i = v2.begin(); i != v2.end(); ++i){
		cout << *i << " ";
	}
    cout << endl << endl;

    // Pomoci iteratoru lze pro zakladni kontejnerove objekty ze STL
    // menit, vkladat a odebirat prvky. Mezi zakladni typy patri
    // vector, list, deque.

    cout << "--- Vypsani hodnoty z pozice 4 vektoru v2 iteratorem ---" << endl;
    vector<int>::iterator i = v2.begin();
	i+=3;
	cout << *i << endl << endl;

	cout << "--- Vlozeni hodnoty doprostred vektoru v2 ---" << endl << endl;

	v2.insert(v2.begin() + v2.size()/2, 1000);
	for(int i = 0; i < (int) v2.size(); i++){
		cout << v2[i] << " ";
	}
	cout << endl << endl;

	cout << "--- Odebrani hodnoty ze stredu vektoru v2 ---" << endl;
	v2.erase(v2.begin() + v2.size() / 2);
	for(int i = 0; i < (int) v2.size(); i++){
		cout << v2[i] << " ";
	}
	cout << endl << endl;

    return 0;
}
